(function () {
    'use strict';

    /**
     * Logger that exposes the same API to master and to worker.
     * When used in a worker context, it will bubble the message to the master process.
     * When used in a master context, it will pipe the message to npmlog.
     * npmlog may then have its stream pointed to anything useful, or be listened to for specific
     * events.
     */

    var cluster = require('cluster');
    var npmlog = require('npmlog');
    var logger;

    if (cluster.isMaster) {
        // Master logger is npmlog, with an extra "access" canal for access logs.
        logger = npmlog;

        var messageHandler = function (msg) {
            logger.log(msg.level, msg.prefix, msg.message);
        };

        var forkHandler = function (worker) {
            worker.on('message', messageHandler);
        };

        logger.hook = function () {
            cluster.on('fork', forkHandler);
        };

        logger.unhook = function () {
            cluster.removeListener('fork', forkHandler);
        };

    } else {
        logger = new (function () {
            this.log = function (level, prefix, message) {
                try {
                    process.send({
                        level: level,
                        prefix: (prefix ? (prefix + ' ') : '') + '(worker #' + process.pid  + ')',
                        message: message
                    });
                } catch (e) {
                    npmlog.log(level, prefix, message);
                    npmlog.error('Logger', 'Failed writing to master process btw');
                }
            };

            ['silly', 'verbose', 'info', 'http', 'warn', 'error'].forEach(function (level) {
                this[level] = this.log.bind(this, level);
            }, this);
        })();
    }

    /**
     * This part is about access handling
     */
    if (cluster.isMaster) {
        // On master, create a new level, that doesn't show-up on the main stream
        logger.addLevel('access', -Infinity);
        // Whenever we get events, and if a writable stream is configured, write
        logger.on('log.access', function (data) {
            if (this.accessStream) {
                this.accessStream.write(data.message + '\n');
            }
        });
    } else {
        // A worker - declare an access method that serializes stuff
        var months = [
            'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
            'Jul',  'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
        ];

        var addDigit = function (n) {
            if (n < 10) {
                return '0' + n;
            }
            return n;
        };

        logger.access = function (data) {
            var line = '', date = new Date(data.currentTime);
            // Remote addr
            if (!data.remoteAddr || data.remoteAddr.slice(0, 2) !== '::') {
                line += '::ffff:';
            }
            line += data.remoteAddr;
            // Empty
            line += ' - - ';
            // Date
            line += '[';
            line += addDigit(date.getUTCDate());
            line += '/';
            line += months[date.getUTCMonth()];
            line += '/';
            line += date.getFullYear();
            line += ':';
            line += addDigit(date.getUTCHours());
            line += ':';
            line += addDigit(date.getUTCMinutes());
            line += ':';
            line += addDigit(date.getUTCSeconds());
            line += ' +0000] "';
            // Request
            line += data.method;
            line += ' ';
            line += data.url;
            line += ' HTTP/';
            line += data.httpVersion;
            line += '" ';
            // Status code
            line += data.statusCode;
            line += ' ';
            // Bytes sent
            //FIXME, sometimes we cannot read socketBytesWritten (maybe because of a websocket?)
            line += data.socketBytesWritten || 0;
            line += ' "';
            // Referer
            line += data.referer || '';
            line += '" "';
            // User-Agent
            line += data.userAgent || '';
            line += '" "';
            // Virtual host
            line += data.virtualHost;
            line += '" ';
            // Total time spent
            line += (data.totalTimeSpent / 1000);
            line += ' ';
            // Backend time spent
            line += (data.backendTimeSpent / 1000);

            this.log('access', '', line);
        };
    }

    module.exports = logger;

})();
